#runnable_blinds.py
import time
from .blinds_circuit import BlindsCircuit
import pandas as pd
#import swifter

class RunnableBlinds():

    def __init__(self):
        self.file = "./data/rolete_seznam_.csv"
        self.my_blinds = BlindsCircuit()


        self.blinds_csv = pd.read_csv(self.file)
        self.filter_blinds = (lambda column, value: self.blinds_csv[self.blinds_csv[column] == value]['id'].tolist())     #.tolist()
        print(f"blinds by smer:{self.filter_blinds('smer', 'zahod')}")
        self.blind_controls = self.my_blinds.get_blind_controls_list()
        self.blinds = self.my_blinds.get_blinds_list()
        #self.my_blinds.set_blinds("down", [1,2,3,6,8,24])

    def run(self):
        while 1:
            time.sleep(0.5)
            self.my_blinds.mcp[4].read_input_on_interrupt()
            interrupted_blind_control = ""
            if(self.my_blinds.mcp[4].bits_gpa.any(1) or self.my_blinds.mcp[4].bits_gpb.any(1)):
                start_time = time.clock_gettime
                for x in self.my_blinds.mcp[4].blind_controls:
                    #print(f"x.blind_number:{x.get_input()} {self.my_blinds.mcp[4].bits_gpa}")
                    if(x.get_input() != None):
                        interrupted_blind_control = x
                        self.my_blinds.set_blinds(blinds_direction=x.get_input()[1], blinds_controlled = self.filter_blinds("etaza", x.blind_number+1))
                        print(f"x.blind_number:{x.get_input()} gpa:{self.my_blinds.mcp[4].bits_gpa} gpb:{self.my_blinds.mcp[4].bits_gpb}")
                        self.my_blinds.apply_changes_to_output()
                        time.sleep(40)
                        
                        self.my_blinds.set_blinds(blinds_direction="stop", blinds_controlled = self.filter_blinds("etaza", x.blind_number+1))
                        self.my_blinds.apply_changes_to_output()

                    
   
#runnable = RunnableBlinds()


import time
from .mcp_expander import McpExpander, Blind, BlindControl
import RPi.GPIO as GPIO


class BlindsCircuit():
    mcp = []
    blinds_list = []
    blind_controls_list = []
    
    def __init__(self):
        GPIO.setwarnings(False)
        GPIO.setmode(GPIO.BCM)
        GPIO.setup(22, GPIO.OUT)
        GPIO.output(22, GPIO.LOW)
        time.sleep(.5)
        GPIO.output(22, GPIO.HIGH)
        GPIO.setup(23, GPIO.IN)
        GPIO.add_event_detect(23, GPIO.RISING, callback=self.event_detect, bouncetime=300)

        for value in range(0x20, 0x25):
            self.mcp.append(McpExpander(address=value,))

        
        for l in self.mcp:
            if(isinstance(l, Blind)):
                self.blinds_list.append(l.get_blinds_list())
            elif(isinstance(l, BlindControl) ):
                self.blind_controls_list.append(l.get_blinds_control_list())

    """ function called from GPIO.add_event_detect"""
    def event_detect(self):
        print("not working my callback from event detect")

    """get unnested list of blinds"""
    def get_blinds_list(self): 
        return_list = []
        for x in self.mcp :
            for blind in x.get_blinds_list():
                return_list.append(blind)
        return return_list

    """get unnested list of blind controls (input switches)"""
    def get_blind_controls_list(self): 
        return_list = []
        for x in self.blind_controls_list:
            for blind_controls in x:
                return_list.append(blind_controls)
        return return_list

    def set_blinds_up(self, blinds_direction = "up", blinds_controlled = [], *args, **kvargs):
        self.set_blinds(blinds_direction="up", blinds_controlled = blinds_controlled)

    def set_blinds_down(self, blinds_direction = "down", blinds_controlled = [], *args, **kvargs):
        self.set_blinds(blinds_direction="down", blinds_controlled = blinds_controlled)

    def set_blinds(self, blinds_direction = "stop", blinds_controlled = [], *args, **kvargs):
        #print(f"self.get_blinds_list()):{self.get_blinds_list())}, blinds_controlled:{blinds_controlled}")
        for i in blinds_controlled:
            if(i < len(self.get_blinds_list())):
                #blind = self.get_blinds_list()[i]
                #getattr(blind, (f"{(blinds_direction).lower()}"))()
                getattr(self.get_blinds_list()[i],  (f"{(blinds_direction).lower()}"))()
                #exec(f"self.get_blinds_list(){[i]}.{(blinds_direction).lower()}()")

            


    def apply_changes_to_output(self):
        for m in self.mcp:
            m.apply_changes_to_output()

    def get_interrupt_signal(self):
        if(GPIO.input(23)):
            self.mcp[4].read_input_on_interrupt()
        return GPIO.input(23)


    def __str__(self):
        return_string = ""
        for s in self.mcp:         
            return_string = "#{} {}#\n".format(return_string, str(s) )
        return return_string

    



#from blind import blind
from bitstring import BitArray
from smbus2 import SMBus 
import RPi.GPIO as GPIO

address_map = {
    0x00: 'IODIRA',   0x01: 'IODIRB',   0x02: 'IPOLA',   0x03: 'IPOLB',
    0x04: 'GPINTENA', 0x05: 'GPINTENB', 0x06: 'DEFVALA', 0x07: 'DEFVALB',
    0x08: 'INTCONA',  0x09: 'INTCONB',  0x0a: 'IOCONA',   0x0b: 'IOCONB',
    0x0c: 'GPPUA',    0x0d: 'GPPUB',    0x0e: 'INTFA',   0x0f: 'INTFB',
    0x10: 'INTCAPA',  0x11: 'INTCAPB',  0x12: 'GPIOA',   0x13: 'GPIOB',
    0x14: 'OLATA',    0x15: 'OLATB'
}
    
register_map = {value: key for key, value in address_map.items()}
#max_len = max(len(key) for key in register_map)

class Blind():
    blind_count = []
    blind_count.append(0)
    

    def __init__(self, *args, **kwargs):
        self.address = 0x20
        self.gpio_register = "gpa"
        self.interrupt_bits = "empty"
        self.output_bits = "empty"
        self.down_bit = 0
        self.blind_number = self.blind_count[0]
        self.blind_count[0] = self.blind_count[0] +1
        
        for key, value in kwargs.items():
            for var in vars(self):  #iterate through vriables in function
                if(key == var):
                    setattr(self, var, value)  #assign value to function variable
                
        self.up_bit = self.down_bit + 1

    #def inc_blind_count
    def up(self):
        if(self.interrupt_bits.all(1, [self.down_bit, self.up_bit]) and self.output_bits.all(1, [self.down_bit, self.up_bit])):
            self.output_bits.set(False, self.up_bit)
            print(f"Blind: {self.blind_number}, going ↑UP↑ ...")
        else:
            self.output_bits.set(True, [self.down_bit, self.up_bit])
            print(f"Blind: {self.blind_number}, ↑stop↑ ...")

    def down(self):
        if(self.interrupt_bits.all(1, [self.down_bit, self.up_bit-1]) and self.output_bits.all(1, [self.down_bit, self.up_bit])):
            self.output_bits.set(False, self.down_bit)
            print(f"Blind: {self.blind_number}, going ↓DOWN↓ ...")
        else:
            self.output_bits.set(True, [self.down_bit, self.up_bit])
            print(f"Blind: {self.blind_number}, ↓stop↓ ...")

    def stop(self):
        if(self.interrupt_bits.any(1, [self.down_bit, self.up_bit]) or self.output_bits.any(1, [self.down_bit, self.up_bit])):
            self.output_bits.set(True, [self.down_bit, self.up_bit])
            print(f"Blind: {self.blind_number}, ←stop→ ...")

    def __str__(self):
        return_string = f"Blind number:{ self.blind_number }, Adress:{self.address}, {self.gpio_register},output_bits:{self.output_bits[self.down_bit: self.up_bit+1]}, " 
        return str(f"{return_string} interrupt_bits:{self.interrupt_bits[self.down_bit: self.up_bit+1]} " )

class BlindControl():
    blind_count = []
    blind_count.append(0)

    def __init__(self, *args, **kwargs):
        self.address = 0x24
        self.gpio_register = "gpa"
        self.interrupt_bits = "empty"
        self.input_bits = "empty"
        self.down_bit = 0
        self.blind_number = self.blind_count[0]
        self.blind_count[0] = self.blind_count[0] +1
        
        for key, value in kwargs.items():
            for var in vars(self):  #iterate through vriables in function
                if(key == var):
                    setattr(self, var, value)  #assign value to function variable
        self.up_bit = self.down_bit + 1


    def get_input(self):
        if(self.input_bits[self.up_bit]):
            print(f"Imput: Switch {self.blind_number}, pushed for ↑UP↑ ...")
            return [self.blind_number, "UP"]
        if(self.input_bits[self.down_bit]):
            print(f"Imput: Switch {self.blind_number}, pushed for ↓DOWN↓ ...")
            return [self.blind_number, "DOWN"]
        else:
            pass


    def __str__(self):
        return_string = f"Blind control number:{self.blind_number}, Adress:{self.address}, {self.gpio_register}, input_bits:{self.input_bits[self.down_bit: self.up_bit+1]}"
        return str(f"{return_string} interrupt_bits:{self.interrupt_bits[self.down_bit: self.up_bit+1]} ")



class McpExpander():

    def __init__(self, *args, **kwargs):
        self.blinds = []
        self.blind_controls = []
        self.address = 0x20
  
        self.output_state = {'gpa': BitArray(uintle=0xff, length=8), 'gpb': BitArray(uintle=0xff, length=8)}
        self.interrupt_state = {'gpa': BitArray(uintle=0xff, length=8), 'gpb' : BitArray(uintle=0xff, length=8)}
        self.input_state = {'gpa': BitArray(uintle=0x00, length=8), 'gpb': BitArray(uintle=0x00, length=8)}

        self.bits_gpa = self.output_state["gpa"]
        self.bits_gpb = self.output_state["gpb"]
        self.bits_gpa_i = self.interrupt_state["gpa"]
        self.bits_gpb_i =self.interrupt_state["gpb"]

        for key, value in kwargs.items():
            """assign values to variables"""
            for var in vars(self):
                if(key == var):
                   setattr(self, key, value)


        self.mcp_expander = {"address": self.address, "output_state": self.output_state, "interrupt_state": self.interrupt_state}

        if (self.address < 0x24):
            self.make_blinds_from_outputs()
        else:
            self.make_blind_controls_from_inputs()

    """make eight blinds from gpa(eight bits) and gpb(gpb eight) bits, each blind takes up two bits"""
    def make_blinds_from_outputs(self, address = 0x20):
        device = self.address
        with SMBus(1) as bus:
            bus.write_byte(0x77, 0x05)
            bus.write_byte_data(device, register_map['GPIOA'], 0xFF)
            bus.write_byte_data(device, register_map['GPIOB'], 0xFF)
            bus.write_byte_data(device, register_map['GPPUA'], 0xFF)
            bus.write_byte_data(device, register_map['GPPUB'], 0xFF)
            bus.write_byte_data(device, register_map['IODIRA'], 0x00)
            bus.write_byte_data(device, register_map['IODIRB'], 0x00)

        for i in range(8):
            down_bit = i*2
            if(i < 4): #gpa bits
                self.blinds.append(Blind (address=self.address, output_bits=self.bits_gpb, 
                    interrupt_bits=self.bits_gpb_i, gpio_register="gpb", down_bit=down_bit))
            elif(i >= 4): #gpa bits
                down_bit -= 8
                self.blinds.append(Blind (address=self.address, output_bits=self.bits_gpa, 
                    interrupt_bits=self.bits_gpa_i, gpio_register="gpa", down_bit=down_bit))


    def make_blind_controls_from_inputs(self, address = 0x24, number_of_blind_controls=8):
        device = self.address
        with SMBus(1) as bus:
            bus.write_byte(0x77, 0x05)
            bus.write_byte_data(0x24, register_map['IOCONA'], 0x6A) #bit7=bank, bit6=mirror,   
            bus.write_byte_data(0x24, register_map['IODIRA'], 0xFF)
            bus.write_byte_data(0x24, register_map['IODIRB'], 0xFF)
            bus.write_byte_data(0x24, register_map['GPINTENA'], 0xFF)
            bus.write_byte_data(0x24, register_map['GPINTENB'], 0xFF)
            bus.write_byte_data(0x24, register_map['DEFVALA'], 0xFF)
            bus.write_byte_data(0x24, register_map['DEFVALB'], 0xFF)
            bus.write_byte_data(0x24, register_map['INTCONA'], 0xFF)
            bus.write_byte_data(0x24, register_map['INTCONB'], 0xFF)
            bus.write_byte_data(0x24, register_map['IPOLA'], 0xFF)
            bus.write_byte_data(0x24, register_map['IPOLB'], 0xFF)


           #value_a = bus.read_byte_data(0x24, register_map['INTCAPA'])
            #value_b = bus.read_byte_data(0x24, register_map['INTCAPB'])
            gpioa = bus.read_byte_data(0x24, register_map['GPIOA'])
            gpiob = bus.read_byte_data(0x24, register_map['GPIOB'])
            self.bits_gpa_i.overwrite(f"0x{gpioa:02x}", 0)
            self.bits_gpb_i.overwrite(f"0x{gpiob:02x}", 0)
            self.bits_gpa.overwrite(f"0x{gpioa:02x}", 0)
            self.bits_gpb.overwrite(f"0x{gpiob:02x}", 0)
        
        for i in range(number_of_blind_controls):
            down_bit = i*2
            if(i < 4): #gpb bits
                self.blind_controls.append(BlindControl (address=self.address, input_bits=self.bits_gpb, 
                    interrupt_bits=self.bits_gpb_i, gpio_register = "gpb", down_bit = down_bit, output_state = self.output_state))
            elif(i >= 4): #gpa bits
                down_bit -= 8
                self.blind_controls.append(BlindControl (address=self.address, input_bits=self.bits_gpa, 
                    interrupt_bits=self.bits_gpa_i, gpio_register = "gpa", down_bit = down_bit, output_state = self.output_state))

    def read_input_on_interrupt(self):
        with SMBus(1) as bus:
            bus.write_byte(0x77, 0x05)
            #vala = bus.read_byte_data(0x24, register_map['GPIOA'])
            #valb = bus.read_byte_data(0x24, register_map['GPIOB'])
            #value_a = bus.read_byte_data(0x24, register_map['INTCAPA'])
            #value_b = bus.read_byte_data(0x24, register_map['INTCAPB'])
            gpioa = bus.read_byte_data(0x24, register_map['GPIOA'])
            gpiob = bus.read_byte_data(0x24, register_map['GPIOB'])
            self.bits_gpa.overwrite(f"0x{gpioa:02x}", 0)
            self.bits_gpb.overwrite(f"0x{gpiob:02x}", 0)



    def apply_changes_to_output(self):
        with SMBus(1) as bus:
            bus.write_byte(0x77, 0x05)
            bus.write_byte_data(self.address, register_map['GPIOA'], self.bits_gpa.uintle)
            bus.write_byte_data(self.address, register_map['GPIOB'], self.bits_gpb.uintle)
            gpioa = bus.read_byte_data(self.address, register_map['GPIOA'])
            gpiob = bus.read_byte_data(self.address, register_map['GPIOB'])
            self.bits_gpa_i.overwrite(f"0x{gpioa:02x}", 0)
            self.bits_gpb_i.overwrite(f"0x{gpiob:02x}", 0)
                

    def get_blinds_list(self):
        return self.blinds

    def get_blinds_controls_list(self):
        return self.blinds_controls


    def __str__(self): 
        return_string = f"McpAddress: {self.address}, {self.output_state}, {self.interrupt_state}\n"
        for s in self.blinds:
                return_string = f"{return_string} {str(s)}\n"
        for s in self.blind_controls:
                return_string = f"{return_string} {str(s)}\n"
        return return_string
